//////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2003, The Institute for Genomic Research (TIGR), Rockville,
// Maryland, U.S.A.  All rights reserved.
//////////////////////////////////////////////////////////////////////////////
#include "GffFactory.h"
#include "GenePredDataSet.h"
#include "SpliceSiteDataSet.h"
#include "HomologyDataSet.h"
#include "MatchPrediction.h"
#include "genemodels.h"
#include "DnaStr.h"
#include <fstream>
#include <stdexcept>

static DataSetAbstractProduct* gffSpliceRead(std::ifstream&,const string&);

static string conv2exon(const string& str) {
  if(strstr(str.c_str(),"final")) return "terminal"; 
  else if(strstr(str.c_str(),"term")) return "terminal"; 
  else if(strstr(str.c_str(),"initial")) return "initial"; 
  else if(strstr(str.c_str(),"internal")) return "internal"; 
  else if(strstr(str.c_str(),"single")) return "single"; 
  else if(strstr(str.c_str(),"first")) return "initial"; 
  else if(strstr(str.c_str(),"last")) return "initial"; 
  else if(strstr(str.c_str(),"start")) return "initial"; 
  else if(strstr(str.c_str(),"stop")) return "terminal"; 
  else return "internal";
}

static string get_tab_elem(const string& str, unsigned ith_tab) {
  string::size_type pos=0,stop=string::npos; 
  if( pos >= str.size() ) {
    std::cerr<<"GffFactory:get_tab_elem failed"<<std::endl;
    std::cerr<<"["<<str<<"]"<<std::endl;
    std::cerr<<pos<<" "<<ith_tab<<std::endl;
    return "";
  }
  for(unsigned i = 0; i <= ith_tab; ++i) {
    stop = str.find("\t",pos);
    if(stop==string::npos)
      stop = str.size();
    if( i < ith_tab ) 
      pos = stop+1;
  }
  string sstr = str.substr(pos,stop-pos);
  return sstr;
}

DataSetAbstractProduct* GffFactory::createDataSetAbstractProduct(const string& filename, const string& type, const string& evidence_type) {
  std::ifstream ifs(filename.c_str());
  if(!ifs)
    throw std::invalid_argument(filename + " is not a valid input file ");
  if(strcmp(evidence_type.c_str(),"spliceprediction")==0) {
    return gffSpliceRead(ifs,type);
  } else if(strcmp(evidence_type.c_str(),"homology")==0) {
    return gffHomologyRead(ifs,type);
  }
  const int buff_size = 4096;
  char buff[buff_size];
  SeqLoc sl(dsu::eEither,-1,-1);
  // create a new list of genes
  GenePredDataSet* newGenePredDataSet = new GenePredDataSet(sl,-1,type);
  int debug_cnt = 0, geneId = 0;
  string geneIdStr,lastGeneIdStr;
  string gffid = type;
  while( ifs.getline(buff,buff_size) ) {
    if( *buff == '#' || *buff == '\n' || !*buff ) 
      continue;
    ++debug_cnt;
    string exon_type = get_tab_elem(buff,2);
    exon_type = conv2exon(exon_type);
    int end5 = atoi(get_tab_elem(buff,3).c_str());
    int end3 = atoi(get_tab_elem(buff,4).c_str());
    string buff_five = get_tab_elem(buff,5);
    double score = 1.0;
    if( buff_five != string(".") ) {
      score = atof(buff_five.c_str());
    }
    string geneIdStr = get_tab_elem(buff,8);
    if( geneIdStr != lastGeneIdStr ) {
      ++geneId;
    }
    string strnd = get_tab_elem(buff,6);
    dsu::Strand_t parseStrnd = dsu::eEither; 
    if( strnd == "-" ) {
      parseStrnd = dsu::eNeg;
    } else {
      parseStrnd = dsu::ePos;
    }
    try {
      assert(end5 >= 0 && end3 >= 0);
      SeqLoc theSl(parseStrnd,end5,end3);
      newGenePredDataSet->insert(gffid,geneId,theSl,exon_type,score);
    } catch(std::exception& e) {
      std::cerr<<e.what()<<std::endl;
      std::cerr<<"attempting to continue: ["<<filename<<"]"<<std::endl;
    }
    lastGeneIdStr = geneIdStr;
  }
  if(!debug_cnt) {
    throw std::invalid_argument(filename + " : unable to read file");
  }
  return newGenePredDataSet;
}


DataSetAbstractProduct* gffSpliceRead(std::ifstream& ifs, const string& type) {
  SeqLoc sl(dsu::eEither,-1,-1);
  SpliceSiteDataSet* newSpliceSiteDataSet = new SpliceSiteDataSet(sl,-1,type);
  int debug_cnt = 0;
  const int buff_size = 4096;
  char buff[buff_size];
  while( ifs.getline(buff,buff_size) ) {
    if( *buff == '#' || *buff == '\n' ) 
      continue;
    string gffid = get_tab_elem(buff,0)+ string(" ") + get_tab_elem(buff,1);
    string splice_type = get_tab_elem(buff,2);
    int end5 = atoi(get_tab_elem(buff,3).c_str());
    int end3 = atoi(get_tab_elem(buff,4).c_str());
    double score = atof(get_tab_elem(buff,5).c_str());
    string strnd = get_tab_elem(buff,6);
    if(strnd == "-") {
      std::swap(end5,end3);
    }
    ++debug_cnt;
    try {
      assert(end5 >= 0 && end3 >= 0);
      newSpliceSiteDataSet->insert(type, end5,end3,splice_type,score);
    } catch(std::exception& e) {
      std::cerr<<e.what()<<std::endl;
      std::cerr<<"attempting to continue..."<<std::endl;
    }
  }
  if(!debug_cnt) 
    throw std::logic_error(type + " unable to get any data from this file");
    
  newSpliceSiteDataSet->begin();
  return newSpliceSiteDataSet;
}


DataSetAbstractProduct* GffFactory::gffHomologyRead(std::ifstream& ifs, const string& type) {
  int debug_cnt = 0;
  HomologyDataSet* newHomologyDataSet = new HomologyDataSet(SeqLoc(dsu::eEither,-1,-1),type);
  const int buff_size = 2024;
  char buff[buff_size];
  double native_score =-1,percsim=-1,lscore=-1;
  while( ifs.getline(buff,buff_size) ) {
    if( *buff == '#' || *buff == '\n' ) 
      continue;
    ++debug_cnt;
    string accession = get_tab_elem(buff,0)+ string(" ") + get_tab_elem(buff,1);
    string exon_type = get_tab_elem(buff,2);
    int end5 = atoi(get_tab_elem(buff,3).c_str());
    int end3 = atoi(get_tab_elem(buff,4).c_str());
    double pcid = atof(get_tab_elem(buff,5).c_str());
    string strnd = get_tab_elem(buff,6);
    dsu::Strand_t parseStrnd = dsu::eEither; 
    if( strnd == "-" ) {
      parseStrnd = dsu::eNeg;
    } else {
      parseStrnd = dsu::ePos;
    }
    try {
      dsu::Strand_t theStrand = dsu::eEither;
      if(end5 > end3) {
	std::swap(end5,end3);
	theStrand = dsu::eNeg;
      } else {
	theStrand = dsu::ePos;
      }
      assert(end5 >= 0 && end3 >= 0);
      SeqLoc sl(theStrand,end5,end3);
      SeqLoc rsl( dsu::eEither, -1, -1 );
      MatchPrediction* mp = new MatchPrediction(sl,native_score,-1,accession,pcid,percsim,lscore,rsl,Internal::getInstance());
      newHomologyDataSet->insert(mp,type);
      //** TODO
      //** stop making the genepredata extra indirection
      //** but you have to correct the checkForMatchPrediction function
      //** and then retest
    } catch(std::exception& e) {
      std::cerr<<e.what()<<std::endl;
      std::cerr<<"attempting to continue..."<<std::endl;
    }
  }
  if(!debug_cnt) 
    throw std::logic_error(type + " unable to get any data from this file");

  return newHomologyDataSet;
}
