//////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2003, The Institute for Genomic Research (TIGR), Rockville,
// Maryland, U.S.A.  All rights reserved.
//////////////////////////////////////////////////////////////////////////////
#include "InputData.h"
#include "EvScores.h"
#include "FormatRegistry.h"
#include "EvidenceTypeRegistry.h"
#include "Options.h"
#include "DataDescrip.h"
#include <strstream>
#include "DataSetAbstractProduct.h"
#include "DnaStr.h"
#include <stdexcept>
#include <fstream>
#include <algorithm>

using std::endl;
using std::cerr;
using std::cout;

InputData::~InputData() {
  for(unsigned i = 0; i < _data.size(); ++i) {
    _data[i]->del();
    delete _data[i];
  }
  list<DataSetAbstractProduct*>::iterator iter = _stopLeak.begin();
  list<DataSetAbstractProduct*>::iterator stop = _stopLeak.end();
  while(iter != stop) {
    DataSetAbstractProduct* dp = *iter;
    dp->del();
    delete dp;
    ++iter;
  }
}

static bool ignoreCaseEqual(const string& a, const string& b); 

static string createUniqId(const string& filename, int id) {
  string::size_type pos = filename.rfind("/");
  if(pos == string::npos) {
    pos = filename.rfind("\\");
  }
  std::ostrstream ostrm;
  ostrm<<id<<'\0';
  string retval = filename.substr( pos==string::npos?0:pos+1,filename.size()) + ostrm.str();
  return retval;
}

static string getStr(const string& str, string::size_type& pos) {
  if(pos >= str.size()) {
    cerr<<"error parsing string: ["<<str<<"]"<<endl;
    cerr<<"Unable to continue, did you forgot to tell me what type of evidence this is?"<<endl<<endl<<endl;
    exit(1);
  } 
  string::size_type stop = str.find(" ",pos);
  if(stop==string::npos)
    stop = str.size();
  string ret = str.substr(pos,stop-pos);
  pos = stop+1;
  return ret;
}

static string int2str(int idx) {
  switch(idx) {
  case EvScores::eAcc: return "acc";
  case EvScores::eDon: return "don";
  case EvScores::eCoding: return "coding";
  case EvScores::eStart: return "start";
  case EvScores::eStop: return "stop";
  default: return "error";
  }
}

static double getDouble(const string& str, string::size_type& pos) {
  const Options& opt = *(Options::getInstance());
  bool linComb = opt.isLinComb();
  if( !linComb ) {
    return 1.0;
  }
  if( pos >= str.size() ) {
    cerr<<"Unable to parse data file: ["<<str<<"] (getDouble)\n";
    exit(1);
  }
  assert(pos < str.size());
  string::size_type stop = str.find(" ",pos);
  if(stop==string::npos)
    stop = str.size();
  string sstr = str.substr(pos,stop-pos);
  std::istrstream strm(sstr.c_str());
  double retVal = -1;
  strm>>retVal;
  pos = stop+1;
  return retVal;
}

struct Compare : public std::binary_function<const SeqLoc*, const SeqLoc*,bool> {
  bool operator()(const SeqLoc* a, const SeqLoc* b) const {
    return *a < *b;
  }
};

static void local_insert(DataSetAbstractProduct* data, DataSetAbstractProduct& product) {
  Prediction::DataStorageType::iterator iter = data->begin();
  while( iter != data->end() ) {
    Prediction::DataStorageType* my_list = (*iter)->makePlaceHolders();
    Prediction::DataStorageType::iterator inner_iter = my_list->begin();
    while( inner_iter != my_list->end() ) {
      product.insert( *inner_iter );
      ++inner_iter;
    }
    delete my_list;
    ++iter;
  }
}

InputData::InputData(const string& filename, const string& theSeq, vector<DataDescrip*>& fms) {
  DataSetAbstractProduct product(SeqLoc(dsu::eEither,-1,-1),-1,filename,AnnotationScore::badVal()); 
  // file contains a list of filenames
  // that hold the actual annotation data
  std::ifstream ifs(filename.c_str());
  if(!ifs) {
     throw std::invalid_argument(filename + " : not a valid filename");
  }
  DataSetAbstractFactory* factory = NULL;
  const int BUFF_SIZE = 15000;
  char buff[BUFF_SIZE];
  // loop through each file
  int numDataSources = 0;
  while(ifs.getline(buff,BUFF_SIZE)) {
    if( *buff == '#') 
      continue;
    string::size_type pos = 0;
    string parseFileName = getStr(buff,pos);
    string format = getStr(buff,pos);
    string type = getStr(buff,pos);
    string usid = createUniqId(parseFileName,numDataSources); 
    vector<EvidenceType*> evd(fms.size());
    for(unsigned i = 0; i < evd.size(); ++i) {
      evd[i] = NULL;
    }
    double wght = getDouble(buff,pos);
    // determine data type by extension for now
    if( ignoreCaseEqual(type,"spliceprediction") ) {
      const int uniqId = fms[EvScores::eAcc]->size();
      evd[EvScores::eAcc] = EvidenceTypeRegistry::createNewEvidenceTypeLinearBackwardCompatible("acc",type,wght,usid,uniqId,theSeq);
      const int uniqId1 = fms[EvScores::eDon]->size();
      evd[EvScores::eDon] = EvidenceTypeRegistry::createNewEvidenceTypeLinearBackwardCompatible("don",type,wght,usid,uniqId1,theSeq);
    } else if( ignoreCaseEqual(type,"homology") || ignoreCaseEqual(type,"geneprediction") ) {
      for(unsigned j = 0; j < fms.size(); ++j) {
	const int uniqId = fms[j]->size();
	const string evType = int2str(j);
	evd[j] = EvidenceTypeRegistry::createNewEvidenceTypeLinearBackwardCompatible(evType,type,wght,usid,uniqId,theSeq);
      }
    } 
    else if( ignoreCaseEqual(type,"curation") ) {
    }  else {
      cerr<<"Don't recognize the "<<type<<" data id type, skipping"<<endl;
      continue;
    }
    factory = FormatRegistry::getFactory(format,type);
    if(factory) {
      try {
	bool someFound = false;
	  for(unsigned i = 0; i < fms.size(); ++i) {
	    if(!evd[i] && type != "curation") {
	      //cerr<<"warning: ["<<type<<"] not described, wont be used in a feature vector"<<endl;
	    } else {
	      someFound = true;
	      DataDescrip& dd = *(fms[i]);
	      if( evd[i] )
		dd.add(evd[i]);
	   }
	}
	if( !someFound ) {
	  cerr<<"warning: ["<<type<<"] not described, wont be used in a feature vector"<<endl;
	}
	DataSetAbstractProduct * data = factory->createDataSetAbstractProduct(parseFileName,usid,type);
	local_insert( data , product);
	_stopLeak.push_back(data);
	++numDataSources;
      } catch(std::exception& e) {
	cerr<<e.what()<<endl;
	cerr<<"Attempting to continue..."<<endl;
      }
    }
  }
  if( !numDataSources ) {
    cerr<<"fatal error, no input data sources read, check the input file: ["<<filename<<"]\n";
    exit(1);
  }
  _data.resize( product.size() );
  DataSetAbstractProduct::iterator iter = product.begin();
  DataSetAbstractProduct::iterator end = product.end();
  int cnt = 0;
  while( iter != end ) {
    _data[cnt] = *iter;
    ++iter;
    ++cnt;
  }
  vector<Prediction*>::iterator start = _data.begin();
  vector<Prediction*>::iterator stop = _data.end();
  std::sort<vector<Prediction*>::iterator,Compare>( start, stop, Compare() );
}

bool ignoreCaseEqual(const string& a, const string& b) {
  for(unsigned i = 0; i < a.size(); ++i) {
    if(tolower(a[i]) != b[i]) return false;
  }
  return true;
}

std::ostream& operator<<(std::ostream& os, const InputData& id) {
  os<<"implment print";
  return os;
}

const string singleStr = "All Input Data";
const string& InputData::getTypeStr() const {
  return singleStr;
}
